source("01-packs-options.R")
#source("02-preproc.R")
source("05-rsaga-dem.R")
source("10-functions-shapefile.R")

# ideas for average linkage clustering
# all functions for average linkage clustering are
# prefixed with "avr"

# triangulation library
library(tripack)

# for debugging purposes
options(digits=2)

prepareClusters <- function(sppdf, framename, numclust, predictors){

        numclust = numclust

        clustering <- makeTessellation(sppdf=sppdf,framename=framename, k=numclust,returncenters=TRUE)

        clf <- list()
        cdf <- list()
        
        clf$data <- sppdf
        
        # fill data frame with clusters and their properties
        for (i in c(1:numclust)){
                # object stores temp. cluster vector (int's of points assigned to cluster)
                cluster <- unlist(clustering[i])
                names(cluster) <- NULL
                cdf[[i]] <- cluster
                names(cdf)[i] <- paste(i)
        }
        
        clf$clustering <- clustering
        clf$clusters <- cdf

        # construct normalised data, needed for unskewed distance calculations
        clf$data.norm <- clf$data
        for (i in predictors){
                clf$data.norm[[i]] <- scale(clf$data[[i]])
        }

        return(clf)
}

generateClusterPrototypes <- function(clf,measure,predictors){

        numclust <- length(clf$clusters)

        if(measure == "average"){
                prototypes.average <- list()

                for(i in 1:numclust){
                        prototypes.average[[i]] <- getClusterPrototype(clf, cluster = i, predictors, measure = "average")
                        names(prototypes.average)[i] <- paste(i)
                }
        }
        clf$prototypes.average <- prototypes.average
        return(clf)
}

addClusterNeighbors <- function(clf,distfact){

        # construct a list of cluster neighbors (one cluster per line, all its direct neighbors per line)
        
        numclust <- length(clf$clusters)
        
        # idea: use delaunay triangulation to determine direct cluster neighbors,
        # the cluster centers which are in one of the delaunay triangles are 
        # direct neighbors

        delaunay <- delaunayn(clf$clustering$clustercenters)
        

        # create list of neighbors for each cluster (indexed by list index number)
        listofneighbors <- list()

        for(i in c(1:numclust)){
               # get indices of rows where the current cluster is found
               ind <- which(delaunay == i, arr.ind=TRUE)[,1] 

               # grab neighbors (actually almost a "select" command)
               uniqueneighborlist <- sort(unique(unlist(as.list(delaunay[ind,]))))

               # remove current cluster from its own list of neighbors
               uniqueneighborlist <- uniqueneighborlist[which(uniqueneighborlist != i)]

               # add list of neighbors for current cluster to "listofneighbors" object
               listofneighbors[[i]] <- uniqueneighborlist
               names(listofneighbors)[i] <- paste(i)
        }

        # remove neighbors which are too far away (distance larger than medium distance)
        # solves a bug/feature in the delaunay triangulation (voronoi tessellation) where 
        # some clusters are neighbors because their borders meet somewhere outside 
        # the actual field borders (this is a _heuristic_, not an exact solution)

        # create distance matrix (dimensions: numclust x numclust)
        distmat <- array(data=NA,dim=c(numclust,numclust))
        cc <- clf$clustering$clustercenters
        for(i in c(1:numclust)){
                for(j in c(1:numclust)){
                distmat[i,j] <- sqrt(crossprod(cc[i,]-cc[j,]))
                }
        }

        # remove neighbors which are too far away (in geo space),
        # heuristic: more than (distfact) distance of nearest neighbor
        # distfact <- distfact
        nb <- listofneighbors

        for(i in names(nb)){
                cn <- nb[i] #current neighbors under consideration
                #namecn <- names(cn)
                #for (c in cn){
                line <- which(names(nb) == i)
                dists <- distmat[line,unlist(cn)]
                mindist <- min(dists)
                keepdists <- dists < distfact*mindist
                nb[[i]] <- cn[[1]][keepdists == TRUE]
        }

        clf$listofneighbors <- listofneighbors
        return(clf)
}

# function operating globally on the whole cluster structure and data frame;
# determines the most similar clusters (with a possible contiguity constraint)

determineClosestClusters <- function(clf){
        
        # clf: cluster structure
        # contigfact: contiguity factor which determines when to switch from the
        #             hard to the soft spatial contiguity constraint
        # (not in this function any more) 2011-01-25

        # constraints for being "most similar":
        # clusters must be spatial neighbors
        # and 
        # clusters must be close in feature space

        # idea (@russ2011icdm,@russ2011scai)
        # -- at the beginning the algorithm enforces spatial contiguity
        #      (i.e. spatial contiguity is a hard constraint)
        # -- towards the end this constraint is relaxed to a soft
        #      constraint (i.e. spatial contiguity is not enforced any more)
        # -- determine the point to switch from the hard to the soft
        #      constraint via the ratio between the median distances in feature space between
        #      a) the current cluster and its spatial neighbors 
        #      b) the current cluster and the other (non-adjacent) clusters
        # if a) gets $contigfact times as high as b), switch to the soft constraint
        # i.e. if (b/a) < contigfact
        # -- to make it more stable (strongly varying medians occur due to the nature of
        #      the algorithm), the moving average over the last few ratios is determined

        # init cluster indices
        clusterswithsmallestdistanceSpatNeigh <- list(0,0)
        clusterswithsmallestdistanceSpatNonNeigh <- list(0,0)
        # init euclid distance
        eucliddistoldspat <- NULL
        eucliddistoldnonspat <- NULL

        for(i in names(clf$clusters)){
                # get current cluster prototype:
                # average vector of vectors in cluster, calculated before
                currentprototype <- unlist(clf$prototypes.average[[i]])

                # grab current spatial neighbors
                currentNeighbors <- clf$listofneighbors[[i]]
#                print("--- current cluster ---")
#                print(i)
#                print("and its currentNeighbors:")
#                print(currentNeighbors)

                # cc: all currentClusters (cluster names)
                cc <- names(clf$clusters)
                currentNonNeighbors <- cc[-match(currentNeighbors,cc)]
                # remove the currently used cluster from its "match"ed nonSpatialNeighbors
                currentNonNeighbors <- currentNonNeighbors[-match(i,currentNonNeighbors)]

                #print("--- current cluster ---")
                #print(i)
                #print("and its currentNonNeighbors:")
                #print(currentNonNeighbors)

                spatialNeighborFeatureSpaceDistances <- vector()
                nonSpatialNeighborFeatureSpaceDistances <- vector()

                # first: stuff for the spatial neighbors
                for(j in currentNeighbors){

                        prototypetocompare <- unlist(clf$prototypes.average[[j]])
                        # determine Euclidean distance in feature space
                        eucliddist <- sqrt(crossprod(currentprototype-prototypetocompare))

                        # store Euclidean distance for the later calculation
                        # of the average distances between spatial neighbors and non-neighbors
                        spatialNeighborFeatureSpaceDistances <- c(spatialNeighborFeatureSpaceDistances,eucliddist)

                        # check if current distance is smaller than the one stored
                        # and update the stored distance

                        # the following condition should be met only once, 
                        # for/at initialisation
                        if (is.null(eucliddistoldspat)){
                                eucliddistoldspat <- eucliddist
                                clusterswithsmallestdistanceSpatNeigh$cl <- list(i,paste(j))        
                                clusterswithsmallestdistanceSpatNeigh$dist <- eucliddist
                        }
                        
                        if (eucliddist < eucliddistoldspat){
                                clusterswithsmallestdistanceSpatNeigh$cl <- list(i,paste(j))        
                                clusterswithsmallestdistanceSpatNeigh$dist <- eucliddist
                                eucliddistoldspat <- eucliddist
                        }
                }

                # second: the same procedure for the spatially non-adjacent 
                # clusters (the "rest")
                for(j in currentNonNeighbors){
                        prototypetocompare <- unlist(clf$prototypes.average[[j]])
                        eucliddist <- sqrt(crossprod(currentprototype-prototypetocompare))
                        #print(paste("eucliddist non neigh", eucliddist, sep=""))
                        nonSpatialNeighborFeatureSpaceDistances <- c(nonSpatialNeighborFeatureSpaceDistances,eucliddist)

                        # check if current distance is smaller than the one stored
                        # and update the stored distance

                        # the following condition should be met only once, 
                        # for/at initialisation
                        if (is.null(eucliddistoldnonspat)){
                                eucliddistoldnonspat <- eucliddist
                                clusterswithsmallestdistanceSpatNonNeigh$cl <- list(i,paste(j))        
                                clusterswithsmallestdistanceSpatNonNeigh$dist <- eucliddist
                        }
                        

                        if ((eucliddist < eucliddistoldnonspat) && (eucliddist > 0)){
                                clusterswithsmallestdistanceSpatNonNeigh$cl <- list(i,paste(j))        
                                clusterswithsmallestdistanceSpatNonNeigh$dist <- eucliddist
                                eucliddistoldnonspat <- eucliddist
                        }
                 }
        }

        #print("spatialNeighborFeatureSpaceDistances")
        #print(spatialNeighborFeatureSpaceDistances)
        spatNeighMedianDistance <- median(spatialNeighborFeatureSpaceDistances)
        #print("spatNeighMedianDistance:")
        #print(spatNeighMedianDistance)

        #print("nonSpatialNeighborFeatureSpaceDistances:")
        #print(nonSpatialNeighborFeatureSpaceDistances)
        spatNonNeighMedianDistance <- median(nonSpatialNeighborFeatureSpaceDistances)
        #print("nonSpatialNeighborFeatureSpaceDistances")
        #print(nonSpatialNeighborFeatureSpaceDistances)
        #print("spatNonNeighMedianDistance:")
        #print(spatNonNeighMedianDistance)

        spatNeighMinDistance <- as.numeric(clusterswithsmallestdistanceSpatNeigh$dist)
        spatNonNeighMinDistance <- as.numeric(clusterswithsmallestdistanceSpatNonNeigh$dist)
        #print("spatNonNeighMinDistance:")
        #print(spatNonNeighMinDistance)

        # return the list of the two clusters with smallest distance and 
        # some more stuff

        returnlist <- list()
        returnlist$closestSpatNeigh <- clusterswithsmallestdistanceSpatNeigh$cl
        returnlist$closestSpatNonNeigh <- clusterswithsmallestdistanceSpatNonNeigh$cl

        returnlist$spatNeighMinDistance <- spatNeighMinDistance
        returnlist$spatNonNeighMinDistance <- spatNonNeighMinDistance
        returnlist$spatNeighMedianDistance <- spatNeighMedianDistance
        returnlist$spatNonNeighMedianDistance <- spatNonNeighMedianDistance
        return(returnlist)
}

# function to merge two clusters in the cluster structure
# 1. works on clf$clusters
# 1.a) merge cluster points into new cluster
# 1.b) delete old two clusters
# 1.c) add newly generated cluster
#
# 2. works on clf$listofneighbors
# 2.a) merge neighbor list into new cluster
# 2.b) delete old clusters
# 2.c) update references to old clusters in the other neighbors lists
#
# 3. works on clf$prototypes.average
# update (regenerate cluster prototypes)
# global update for now, not really efficient
mergeClusters <- function(clf, clustersToMerge, measure, predictors){

        numclust <- length(clf$clusters)

        clustersToMerge <- unlist(clustersToMerge)

        # new cluster index will be one lower than the total number of clusters
        # because: two are removed, one is added
        clnew <- numclust-1

        cl1 <- clustersToMerge[1]
        cl2 <- clustersToMerge[2]

        # 1.a) merge cluster points into new cluster
        newclusterpoints <- c(unlist(clf$clusters[cl1]),unlist(clf$clusters[cl2]))
        newclustername <- paste("(",cl1,",",cl2,")",sep="")
        # 1.b) delete old clusters
        clf$clusters[cl1] <- NULL
        clf$clusters[cl2] <- NULL
        # 1.c) add newly generated cluster
        clf$clusters[[newclustername]] <- as.vector(newclusterpoints)

        #print("## -- neighbors of cl1 --##:")
        #print(clf$listofneighbors[cl1])
        #print("## -- neighbors of cl1 --##:")
        #print(clf$listofneighbors[cl2])

        # 2.a) merge neighbor list into new cluster
        newclusterneighbors <- c(unlist(clf$listofneighbors[cl1]),unlist(clf$listofneighbors[cl2]))
        #print("## -- newclusterneighbors --##:")
        #print(newclusterneighbors)
        newclusterneighbors <- sort(unique(newclusterneighbors))
        #print("## -- newclusterneighbors after sort/unique --##:")
        #print(newclusterneighbors)
        #print("cl1/cl2")
        #print(cl1)
        #print(cl2)
        #print("----")

        # if clusters are spatially adjacent, this also requires removing the
        # names of the merged clusters. First check if cl1/cl2 are in the new
        # cluster's neighbors. If they are, remove them from there.
        
        if(cl1 %in% newclusterneighbors){
                newclusterneighbors <- newclusterneighbors[-which(newclusterneighbors == cl1)]
        }
        if(cl2 %in% newclusterneighbors){
                newclusterneighbors <- newclusterneighbors[-which(newclusterneighbors == cl2)]
        }
        # it may happen that the new cluster appears in its own neighbors list
        # remove it from there * condition removed *
        #if(newclustername %in% newclusterneighbors){
        #        newclusterneighbors <- newclusterneighbors[-which(newclusterneighbors == newclustername)]

        #}
        clf$listofneighbors[[newclustername]] <- newclusterneighbors

        # 2.b) delete old neighbor lists
        clf$listofneighbors[cl1] <- NULL
        clf$listofneighbors[cl2] <- NULL

        # 2.c) update references to old clusters in the other neighbors lists
        for(i in names(clf$listofneighbors)){
                newlist <- clf$listofneighbors[[i]]
                newlist[which(newlist == cl1)] <- newclustername
                newlist[which(newlist == cl2)] <- newclustername
                newlist <- unique(newlist)
                clf$listofneighbors[[i]] <- newlist
        }

        # 3.) update prototypes (delete two old clusters,
        #     calculate prototype for new cluster)

        clf$prototypes.average[cl1] <- NULL
        clf$prototypes.average[cl2] <- NULL

        clf$prototypes.average[[newclustername]] <- getClusterPrototype(clf,newclustername,predictors,measure)

        return(clf)
}

getClusterPrototype <- function(clf,cluster, predictors, measure){
        
        if(measure == "average"){
                # grab points (point indices) in cluster
                currentclusterind <- unlist(clf$clusters[cluster]) 
                currentclusterdata <- data.frame()
                # obtain each predictor's variable's values 
                # (per cluster, see outer loop in generateClusterPrototypes()),
                # put them into data frame.
                for(j in predictors){
                       variablestmp <- clf$data.norm@data[[j]]
                       varcl <- variablestmp[currentclusterind,]
                       currentclusterdata <- rbind(currentclusterdata,varcl)
                }
                # obtain mean (rowMeans), results in prototype vector for cluster
                clustermeans <- rowMeans(currentclusterdata)
                clusterproto <- as.numeric(as.character(clustermeans))

                return(clusterproto)
                }
}

plotClusters <- function(clf, framename, numclust, predictors,filetype, contigfact, number, color, cex){

         k <- length(clf$clusters)

         at <- seq(1,k,1)

         if (color == TRUE){
                # for color clusters
                col <- rainbow(length(at))
         }
         else {
                # for gray-level clusters
                col = gray((c(1:k)*0.8)/k)
                #col = gray(1:(k-1) / k)
         }

         spl <- list()

         colorcounter <- 0

         for(j in names(clf$clusters)) {
              # get indices of points in current cluster
              cc <- clf$clusters[[j]]
              # get according coordinates
				 			if(length(cc) > 1){
              spp <- SpatialPoints(clf$data@coords[cc,1:2])

              # for the spatial layout function in the plot below
              # for each cluster in the partition store the respective
              # points/color in a list
              colorcounter <- colorcounter + 1
              spl[[j]] <- list("sp.points", spp, col = col[colorcounter], pch=19, cex=cex)
							}
              }

        if(filetype == "postscript"){
                postscript(file = paste("results/clusterplots/", framename, 
                "-hsc-cf-",  contigfact, 
                "-", paste(predictors,collapse="-"), 
                "-cl", numclust,
                "-", sprintf("%04d",number),".eps", sep = ""))
                }

        if(filetype == "png"){
                png(file=paste("results/clusterplots/",framename,
                "-hsc-cf-", contigfact, 
                "-", paste(predictors,collapse="-"),
                "-cl", numclust,
                "-", sprintf("%04d",number),".png", sep=""))
                }

        # Make sure there's an "EC25" variable, otherwise the spplot won't display.
        # If there's no EC25 variable, choose a different one and adapt
        # the function below accordingly
        # hint above superseded, just taking the first predictor in "predictors"

        print(
                spplot(clf$data, zcol = predictors[1],
                #do.log=TRUE,
                #key.space=list(x=0, y=0.01, corner=c(0,0)),
                xlab = "Easting", ylab = "Northing",
                main = paste(framename, ", ", k, " zones, ", toString(predictors), sep = ""),
                col.regions = "transparent", 
                sp.layout = spl)
        )
        dev.off()
        }

# below is the MAIN function for the clustering
# which calls most of the subfunctions above

hierarchicalSpatialClustering <- function(sppdf, framename, numclust, measure, predictors, zones, distfact, contigfact, filetype, color, cex){

        clf <- prepareClusters(sppdf,framename, numclust=numclust, predictors=predictors)
        clf <- addClusterNeighbors(clf,distfact)
        clf <- generateClusterPrototypes(clf, measure=measure, predictors = predictors)
        phase <- "hard"
        slidingWindow <- 10

        spatMinDist <- vector()
        nonSpatMinDist <- vector()
        spatMedianDist <- vector()
        nonSpatMedianDist <- vector()

        # either the algorithm switches from the "hard" to the "soft"
        # constraint in its course, then constraintSwitchedAt gets overwritten.
        # Otherwise it never switches and the variable is kept at zero.
        # Only important for the abline in the plots.
        constraintSwitchedAt <- 0

        for(i in c(1:(numclust-zones))){
                print(" ----- beginning of next clustering step -----")
                print(i)
                print(paste("phase: ", phase, sep=""))
                # the function determineClosestClusters returns a result list which
                # is ripped into pieces two lines below
                closestClusters <- determineClosestClusters(clf)
                closestSpatNeigh <- closestClusters$closestSpatNeigh
                closestSpatNonNeigh <- closestClusters$closestSpatNonNeigh

                spatMedianDist <- c(spatMedianDist,closestClusters$spatNeighMedianDistance)
                #print(paste("spatMedianDist: ",spatMedianDist, sep=""))

                # towards the end of the algorithm, there may be no more clusters which
                # are _not_ spatial neighbors for the current cluster. This returns an "NA" value
                # as the spatNonNeighMedianDistance since the current cluster does not have any
                # distances to the (non-existing) non-adjacent clusters. This is handled by
                # the condition below.

                if (!(is.na(closestClusters$spatNonNeighMedianDistance))){
                        nonSpatMedianDist <- c(nonSpatMedianDist,closestClusters$spatNonNeighMedianDistance)
                }
                

                # check if enough measurements of distances are available for 
                # sliding window calculation; and change phase appropriately if 
                # the condition is met

                if (i > slidingWindow){
                        lastSpatMedianDist <- last(SMA(spatMedianDist,n=slidingWindow))
                        #print(paste("lastSpatMedianDist: ",lastSpatMedianDist, sep=""))
                        lastNonSpatMedianDist <- last(SMA(nonSpatMedianDist,n=slidingWindow))
                        #print(paste("lastNonSpatMedianDist: ",lastNonSpatMedianDist, sep=""))
                        #plot(SMA(spatMedianDist,n=10))
                        if((lastNonSpatMedianDist/lastSpatMedianDist < contigfact) &&
                               (!(identical(phase,"soft")))){
                                phase <- "soft"
                                constraintSwitchedAt <- i
                        }
                }

                if (identical(phase, "hard")){
                        # beginning of cluster algorithm
                        clustersToMerge <- closestSpatNeigh
                }else{
                        # towards the end of the clustering

                       if((closestClusters$spatNeighMinDistance > closestClusters$spatNonNeighMinDistance) &&
                                !(is.null(closestClusters$closestSpatNonNeigh))){
                                # the additional condition ensures that non-existent non-neighbors are not
                                # used (if the pair is NULL)
                             clustersToMerge <- closestSpatNonNeigh
                       } else {
                             clustersToMerge <- closestSpatNeigh
                       }
                }

                clf <- mergeClusters(clf, clustersToMerge, measure=measure, predictors = predictors)
                plotClusters(clf, framename, numclust, predictors, filetype, contigfact, number=i, color=color, cex=cex)
        }
        returnlist <- list()
        returnlist$clf <- clf
        returnlist$spatMinDist <- spatMinDist
        returnlist$nonSpatMinDist <- nonSpatMinDist
        returnlist$spatMedianDist <- spatMedianDist
        returnlist$nonSpatMedianDist <- nonSpatMedianDist
        returnlist$constraintSwitchedAt <- constraintSwitchedAt
        return(returnlist)
}


####################
# this function takes the clustering results and creates some figures and
# evaluation from it

plotClusteringResults <- function(clr,framename, contigfact){
        postscript(file=paste("results/",framename, "-cf-", contigfact, "-clustering-ratio-plots-%02d.eps",sep=""), onefile=FALSE,
                        width=6,height=3,paper="special", family="ComputerModern", horizontal=FALSE)
        #################
        # first with ablines at the position of the "hard"->"soft"-switch
        #y <- SMA(clr$nonSpatMedianDist/clr$spatMedianDist,n=10)
        #x <- seq(along=y)
        #plot(x, y)
                #main=paste(framename, "contiguity ratio",sep=", "),
                ##main="contiguity ratio vs. clustering steps",
                #xlab="cluster steps",
                #ylab="moving average (n=10) \n of contiguity ratio",
                #mgp <- c(3,1,2), 
                #type="l")
                #if(!(clr$constraintSwitchedAt == 0)){abline(v=clr$constraintSwitchedAt)}

        par(las=1)

        plot(SMA(clr$nonSpatMedianDist,n=10),
                main=paste(framename, "median distance of non-adjacent clusters",sep=", "),
                xlab="cluster steps",
                ylab="median distance",
        #        par(las=0),
                type="l")
                if(!(clr$constraintSwitchedAt == 0)){abline(v=clr$constraintSwitchedAt)}
        plot(SMA(clr$spatMedianDist,n=10),
                main=paste(framename, "median distance of adjacent clusters",sep=", "),
                xlab="cluster steps",
                ylab="median distance",
        #        par(las=0),
                type="l")
                if(!(clr$constraintSwitchedAt == 0)){abline(v=clr$constraintSwitchedAt)}
        # then without those ablines 
        plot(-(SMA(clr$nonSpatMedianDist/clr$spatMedianDist,n=10)),
                main=paste(framename, "contiguity ratio",sep=", "),
                #main="contiguity ratio vs. clustering steps",
                xlab="cluster steps",
                ylab="contiguity ratio",
                ylim=c(-2.1,-0.5),
        #        par(las=0),
                type="l")
        plot(SMA(clr$nonSpatMedianDist,n=10),
                main=paste(framename, "median distance of non-adjacent clusters",sep=", "),
                xlab="cluster steps",
                ylab="median distance",
        #        par(las=0),
                type="l")
        plot(SMA(clr$spatMedianDist,n=10),
                main=paste(framename, "median distance of adjacent clusters",sep=", "),
                xlab="cluster steps",
                ylab="median distance",
        #        par(las=0),
                type="l")

        # unsmoothed
        plot(-clr$nonSpatMedianDist/clr$spatMedianDist,
                main=paste(framename, "contiguity ratio",sep=", "),
                xlab="cluster steps",
                ylab="ratio",
                ylim=c(-3,0),
        #        par(las=0),
                type="l")
        # non-smoothed and smoothed values in one diagram

        #nSMD <- clr$nonSpatMedianDist
        # clip spatMedianDist at the end, since nSMD is necessarily shorter
        #sMD <- clr$spatMedianDist[1:length(nSMD)]

        #par(las=1, xpd=TRUE,mar=par()$mar+c(0,0,0,8))

        #matplot(y=cbind(-(nSMD/sMD), SMA(-(nSMD/sMD),n=10)),
        #        #pch=21:22,
        #        type="l",
        #        col=1,
        #        xlab="cluster steps",
        #        ylab="ratio",
        #        ylim=c(-3,0),
        #        main=paste(framename, ", contiguity ratio", sep=""),
        #        )
        #        #legend(200,-1, c("unsmoothed","smoothed"),lty=1:2)

        dev.off()

}
